#!/usr/bin/env python3

"""
Read and cache APSecret KeyVault secrets and keys.
"""

from http.client import HTTPSConnection
import sys
import os
import stat
import ssl
import subprocess
import json

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon.keyvaulthelper import KeyVaultHelper
from apsdk import apenv


class APSecretClient:
    APSECRET_SECRET_DIR = "secrets"
    SSTOOL_PATH = "/ap/app/aptools/sstool"
    ENDPOINT_TIMEOUT = 5
    APSECRET_RETRIES = 2
    encrypted_file_paths = {
        "ACRConnectionData": "/ap/data/apsecrets/acr-conn-data.txt.encr",

        "Geneva-HotPath-Cert": "/ap/data/apsecrets/mdm-cert-prod.pem.encr",
        "Geneva-HotPath-Key": "/ap/data/apsecrets/mdm-key-prod.pem.encr",
        "GenevaINT-HotPath-Cert": "/ap/data/apsecrets/mdm-cert-int.pem.encr",
        "GenevaINT-HotPath-Key": "/ap/data/apsecrets/mdm-key-int.pem.encr",

        "Geneva-GCS-Certificate": "/ap/data/apsecrets/gcs-cert-prod.pem.encr",
        "Geneva-GCS-Key": "/ap/data/apsecrets/gcs-key-prod.pem.encr",
        "GenevaINT-GCS-Certificate": "/ap/data/apsecrets/gcs-cert-int.pem.encr",
        "GenevaINT-GCS-Key": "/ap/data/apsecrets/gcs-key-int.pem.encr"
    }

    ap_key_rep_vips = ["CH01", "CO1C", "BJ1", "BN01", "Bn1", "Bn2", "Bn4c", "Ch1", "Ch1b", "CO01",
                       "Co3", "Co3b", "co3c", "CO4", "CO4C", "cy2", "CY2B", "CY4", "Db3", "Db4",
                       "db5", "dm2c", "Dm3", "HK2", "Sg1", "Sn2", "MW1", "BN2B", "Ch1d", "DUB01",
                       "DUB02", "MWH01", "HKG01", "CY1Prod", "CO1Prod", "CO1bProd", "BN1Prod", "CY1PPE",
                       "BN1Lab", "CHI02"]

    def __init__(self,
                 client_cert_file,
                 client_key_file,
                 endpoints,
                 apsecret_cache="/var/cache/apsecret"):
        self.client_key_file = client_key_file
        self.client_cert_file = client_cert_file
        self.apsecret_endpoints = endpoints
        self.apsecret_cache = apsecret_cache

        self.keyvault_config = ""
        self.tenant = ""
        self.client_id = ""
        self.client_key = ""
        self.vault_uri = ""
        self.cluster = self.get_cluster()

    def get_cluster(self):
        try:
            return apenv.get_cluster_name()

        except Exception as e:
            print("ERROR: Unable to get Cluster from apenv: {0}".format(e), file=sys.stderr)

        return ""

    def is_connection_data_initialized(self) -> bool:
        """
        Checks to see if we have connection data needed to access KeyVault.
        :return: Returns True if the KeyVault connection info is present and otherwise False.
        """
        return bool(self.tenant and self.client_id and self.client_key and self.vault_uri)

    def __get_keyvault_connection_json(self):
        """
        Gets the connection data needed to connect to APLinux Prod's KeyVault instance for the
        purpose of accessing bootstrap secrets.
        """
        if self.client_key_file is None or self.client_cert_file is None:
            raise ValueError("Client certificate is required")

        # configure certificate verification
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        # retrieve the json config
        eps = self.apsecret_endpoints
        if type(eps) is str:
            endpoints = [eps]
        elif type(eps) is list:
            endpoints = eps
        else:
            raise ValueError("Endpoint parameter must be a str or a list of str")

        for endpoint in endpoints:
            for retry in range(0, self.APSECRET_RETRIES):
                try:
                    conn = HTTPSConnection(endpoint,
                                           key_file=self.client_key_file,
                                           cert_file=self.client_cert_file,
                                           context=ctx,
                                           timeout=self.ENDPOINT_TIMEOUT)
                    conn.putrequest("GET", "/keyvault/aplinux")
                    conn.endheaders()
                    response = conn.getresponse()
                except ssl.SSLError as e:
                    if str(e).startswith("[SSL] PEM lib"):
                        raise NotImplementedError("Invalid client cert format")
                    print("ERROR: SSL error connecting to the APSecret service "
                          "endpoint '{0}': {1}. Trying next endpoint...".format(endpoint, e),
                          file=sys.stderr)
                    continue
                except Exception as e:
                    print("ERROR: Unable to connect to the APSecret service "
                          "endpoint '{0}': {1}. Trying next endpoint...".format(endpoint, e),
                          file=sys.stderr)
                    continue

                if response.status != 200:
                    print("WARNING: APSecret service endpoint '{0}' returned with "
                          "an error: {1} {2}. Trying next endpoint...".format(endpoint,
                                                                              response.status,
                                                                              response.reason),
                          file=sys.stderr)
                    continue

                try:
                    data = response.read()
                    conn.close()
                except Exception as e:
                    print("ERROR: Connection not available to APSecret service "
                          "endpoint '{0}': {1}. Trying next endpoint...".format(endpoint, e),
                          file=sys.stderr)
                    continue

                # successfully connected
                if data:
                    break

        if not data:
            raise ConnectionError("Unable to connect to any working APSecret service endpoints")

        self.keyvault_config = data.decode(encoding="UTF-8")
        self.__parse_keyvault_connection_json()

    def __parse_keyvault_connection_json(self):
        """
        Parses the json received from APSecret containing the KeyVault connection information.
        """
        # parse the json config file
        try:
            config = json.loads(self.keyvault_config)
        except:
            raise ValueError("Connection information retrieved from authority is in a unsupported format.")

        if config is None or len(config) == 0:
            raise ValueError("No connection information available from authority.")
        if [o for o in ("TenantID", "ClientID", "ClientSecret", "VaultURI") if o not in config[0]]:
            raise ValueError("Connection information retrieved from authority is incomplete.")

        self.tenant = config[0]["TenantID"]
        self.client_id = config[0]["ClientID"]
        self.client_key = config[0]["ClientSecret"]
        self.vault_uri = config[0]["VaultURI"]

        if not self.is_connection_data_initialized():
            raise ValueError("Connection information retrieved from authority is invalid.")

    def __create_directory_layout(self):
        """
        Creates the directory structure for the cache.
        """
        if not os.path.isdir(self.apsecret_cache):
            try:
                os.makedirs(self.apsecret_cache, 0o750, exist_ok=True)
            except Exception as e:
                raise IOError("Unable to create directory layout of cache: mkdir recursive {0}: {1}".format(self.apsecret_cache, e))

        subdirectory = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR)
        try:
            if not os.path.isdir(subdirectory):
                os.mkdir(subdirectory, 0o750)
        except Exception as e:
            raise IOError("Unable to create directory layout of cache: mkdir {0}: {1}".format(subdirectory, e))

    def __secure_cache_file(self, filepath: str):
        """
        Changes the permissions of the secret file to protect it from unauthorized access.
        """
        try:
            os.chmod(filepath, stat.S_IRWXU | stat.S_IRGRP)
        except Exception as e:
            raise IOError("Failed secure permissions of cache file {0}: {1}".format(filepath, e))

    def get_keyvault_secret(self, name: str) -> str:
        """
        Connects to KeyVault to retrieve the specified secret. Returns the secret.
        :return: Returns the secret value as a string.
        """
        if not name:
            raise ValueError("Get keyvault secret: Empty name")

        kv = KeyVaultHelper(self.tenant, self.client_id, self.client_key, self.vault_uri)

        try:
            kv.connect()
        except Exception as e:
            raise IOError("Connection failed to KeyVault: {0}".format(e))

        try:
            secret = kv.get_secret(name)
        except Exception as e:
            raise IOError("Failed to get secret '{0}' from KeyVault: {1}".format(name, e))

        return secret

    def __get_encr_filepath(self, name: str):
        """
        Gets the .encr file path from the dictionary defined at the top of the file
        :return: Returns the path of the file as a string.
        """
        if name not in self.encrypted_file_paths:
            raise IOError("Unknown secret name passed {0}".format(name))

        if name in self.encrypted_file_paths:
            filepath = self.encrypted_file_paths[name]
            # Verify encrypted secret file exists
            if not os.path.isfile(filepath):
                raise IOError("Encrypted secret file not found {0} - {1}".format(name, filepath))
            return filepath

    def __decrypt_secret_sstool(self, name: str, target_file: str, cluster: str):
        """
        Decrypts .encr files using sstool
        :return: Returns true or false based on success.
        """
        encr_file = self.__get_encr_filepath(name)

        # Decrypt data running sstool
        # /ap/app/aptools/sstool -d -i <file.encr> -o <file.decr> -p ApKeyRep-Backend.Autopilot.CO4.ap.gbl
        cmd = [self.SSTOOL_PATH,
               '--decrypt',
               '--input', encr_file,
               '--output', target_file,
               '--VIP', "ApKeyRep-Backend.Autopilot.{0}.ap.gbl".format(cluster)
               ]
        cmdp = ' '.join(cmd)
        err = ''

        try:
            p = subprocess.Popen(cmd,
                                 shell=False,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            out, err = p.communicate()
        except Exception as e:
            print("Warning: sstool: Failed to run '{0}': {1}".format(cmdp, e), file=sys.stderr)
            return False

        if p.returncode != 0:
            raise IOError("sstool: Retcode ({0}) running '{1}': {2} : {3}".format(p.returncode,
                                                                                  cmdp,
                                                                                  err.replace('\n', ' ')),
                          out)

        if err:
            raise IOError("sstool: Unexpected error executing {0}: {1}".format(cmdp,
                                                                               err.replace('\n', ' ')))

        return True

    def __update_cache_secret(self, name: str):
        """
        Updates a locally cached secret.
        """
        self.__create_directory_layout()
        target_file = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR, name)
        cluster = self.cluster

        print("NOTICE: Decrypting against original cluster: {0}".format(cluster), file=sys.stderr)

        try:
            # Try decrypting using sstool
            self.__decrypt_secret_sstool(name, target_file, cluster)
            self.__secure_cache_file(target_file)
            return
        except Exception as e:
            print("WARNING: sstool Failed to update cache of secret '{0}': {1}".format(name, e), file=sys.stderr)
            print("WARNING: Trying all other possible clusters before giving up...", file=sys.stderr)

            for cluster in self.ap_key_rep_vips:
                print("WARNING: Trying cluster: {0}".format(cluster), file=sys.stderr)
                try:
                    # Try decrypting using sstool
                    self.__decrypt_secret_sstool(name, target_file, cluster)
                    self.__secure_cache_file(target_file)
                    return
                except Exception as e:
                    print("Warning: sstool: Failed to decrypt against cluster: '{0}': {1}".format(cluster, e), file=sys.stderr)

        # Fall back to keyvault
        if not self.is_connection_data_initialized():
            self.__get_keyvault_connection_json()

        try:
            secret = self.get_keyvault_secret(name)
        except Exception as e:
            raise IOError("Trying Keyvault - Updating secret cache: {0}".format(e))

        # Write the secret to file and secure it
        self.__create_directory_layout()
        filename = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR, name)
        try:
            f = open(filename, "w+")
            f.write(secret)
            f.close()
            self.__secure_cache_file(filename)
        except Exception as e:
            raise IOError("Updating secret cache: Writing secret to cache file {0}: {1}".format(filename, e))

    def read_cache_secret(self, name: str) -> str:
        """
        Gets the value of the specified locally cached secret.
        :return: Returns the secret on success and an exception on error.
        """
        if not name:
            raise ValueError("Read secret cache: Empty name")

        try:
            self.__update_cache_secret(name)
        except Exception as e:
            print("WARNING: Failed to update cache of secret '{0}': {1}".format(name, e), file=sys.stderr)

        filename = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR, name)
        if not os.path.isfile(filename):
            raise KeyError("Read secret cache: '{0}' not cached".format(name))

        try:
            f = open(filename, "r")
            secret = f.read()
            f.close()
        except Exception as e:
            raise IOError("Read secret cache: Unable to read '{0}' from cache: {1}".format(name, e))

        return secret
